@@ -1,21 +1,23 @@
 -----------------------------------------------------------------------------
 -- |
--- Module      :  Data.SBV.Control
--- Copyright   :  (c) Levent Erkok
--- License     :  BSD3
--- Maintainer  :  erkokl@gmail.com
--- Stability   :  experimental
+-- Module    : Data.SBV.Control
+-- Copyright : (c) Levent Erkok
+-- License   : BSD3
+-- Maintainer: erkokl@gmail.com
+-- Stability : experimental
 --
 -- Control sublanguage for interacting with SMT solvers.
 -----------------------------------------------------------------------------
 
-{-# LANGUAGE NamedFieldPuns #-}
+{-# LANGUAGE ConstraintKinds #-}
+
+{-# OPTIONS_GHC -Wall -Werror #-}
 
 module Data.SBV.Control (
      -- $queryIntro
 
      -- * User queries
-       Query, query
+       ExtractIO(..), MonadQuery(..), Queriable(..), Fresh(..), Query, query
 
      -- * Create a fresh variable
      , freshVar_, freshVar
@@ -24,11 +26,11 @@ module Data.SBV.Control (
      , freshArray_, freshArray
 
      -- * Checking satisfiability
-     , CheckSatResult(..), checkSat, checkSatUsing, checkSatAssuming, checkSatAssumingWithUnsatisfiableSet
+     , CheckSatResult(..), checkSat, ensureSat, checkSatUsing, checkSatAssuming, checkSatAssumingWithUnsatisfiableSet
 
      -- * Querying the solver
      -- ** Extracting values
-     , SMTValue(..), getValue, getUninterpretedValue, getModel, getAssignment, getSMTResult, getUnknownReason
+     , SMTValue(..), getValue, registerUISMTFunction, getFunction, getUninterpretedValue, getModel, getAssignment, getSMTResult, getUnknownReason, getObservables
 
      -- ** Extracting the unsat core
      , getUnsatCore
@@ -37,7 +39,7 @@ module Data.SBV.Control (
      , getProof
 
      -- ** Extracting interpolants
-     , getInterpolant
+     , getInterpolantMathSAT, getInterpolantZ3
 
      -- ** Extracting assertions
      , getAssertions
@@ -74,15 +76,33 @@ module Data.SBV.Control (
      , SMTOption(..)
      ) where
 
-import Data.SBV.Core.Data     (SMTConfig(..))
-import Data.SBV.Core.Symbolic (Query, Symbolic, Query(..), QueryContext(..))
+import Data.SBV.Core.Symbolic (MonadQuery(..), Query, Queriable(..), Fresh(..), Symbolic, QueryContext(..))
 
-import Data.SBV.Control.Query
-import Data.SBV.Control.Utils (SMTValue(..), queryDebug, executeQuery)
+import Data.SBV.Control.BaseIO
+import Data.SBV.Control.Query hiding (  getInfo, getOption, getUnknownReason, getObservables
+                                      , getSMTResult, getLexicographicOptResults
+                                      , getIndependentOptResults
+                                      , getParetoOptResults, getModel
+                                      , checkSatAssuming
+                                      , checkSatAssumingWithUnsatisfiableSet
+                                      , getAssertionStackDepth
+                                      , inNewAssertionStack, push, pop
+                                      , caseSplit, resetAssertions, echo, exit
+                                      , getUnsatCore, getProof, getInterpolantMathSAT, getInterpolantZ3
+                                      , getAssertions, getAssignment
+                                      , mkSMTResult, freshVar_, freshVar
+                                      , freshArray, freshArray_, checkSat, ensureSat
+                                      , checkSatUsing, getValue
+                                      , getUninterpretedValue, timeout, io)
+import Data.SBV.Control.Utils (SMTValue, registerUISMTFunction)
+
+import Data.SBV.Utils.ExtractIO (ExtractIO(..))
+
+import qualified Data.SBV.Control.Utils as Trans
 
 -- | Run a custom query
 query :: Query a -> Symbolic a
-query = executeQuery QueryExternal
+query = Trans.executeQuery QueryExternal
 
 {- $queryIntro
 In certain cases, the user might want to take over the communication with the solver, programmatically
@@ -133,7 +153,7 @@ they allow direct control of the solver. Here's a simple example:
 
 Note the type of @test@: it returns an optional pair of integers in the 'Symbolic' monad. We turn
 it into an IO value with the 'Data.SBV.Control.runSMT' function: (There's also 'Data.SBV.Control.runSMTWith' that uses a user specified
-solver instead of the default.)
+solver instead of the default. Note that 'Data.SBV.Provers.z3' is best supported (and tested), if you use another solver your results may vary!)
 
 @
     pair :: IO (Maybe (Integer, Integer))
