@@ -1,17 +1,17 @@
 -----------------------------------------------------------------------------
 -- |
--- Module      :  Data.SBV.Core.AlgReals
--- Copyright   :  (c) Levent Erkok
--- License     :  BSD3
--- Maintainer  :  erkokl@gmail.com
--- Stability   :  experimental
+-- Module    : Data.SBV.Core.AlgReals
+-- Copyright : (c) Levent Erkok
+-- License   : BSD3
+-- Maintainer: erkokl@gmail.com
+-- Stability : experimental
 --
 -- Algrebraic reals in Haskell.
 -----------------------------------------------------------------------------
 
-{-# LANGUAGE FlexibleInstances    #-}
-{-# LANGUAGE TypeSynonymInstances #-}
-{-# OPTIONS_GHC -fno-warn-orphans #-}
+{-# LANGUAGE FlexibleInstances #-}
+
+{-# OPTIONS_GHC -Wall -Werror -fno-warn-orphans #-}
 
 module Data.SBV.Core.AlgReals (
              AlgReal(..)
@@ -22,7 +22,8 @@ module Data.SBV.Core.AlgReals (
            , mergeAlgReals
            , isExactRational
            , algRealStructuralEqual
-           , algRealStructuralCompare)
+           , algRealStructuralCompare
+           )
    where
 
 import Data.List       (sortBy, isPrefixOf, partition)
