@@ -1,43 +1,48 @@
 -----------------------------------------------------------------------------
 -- |
--- Module      :  Data.SBV.Core.Data
--- Copyright   :  (c) Levent Erkok
--- License     :  BSD3
--- Maintainer  :  erkokl@gmail.com
--- Stability   :  experimental
+-- Module    : Data.SBV.Core.Data
+-- Copyright : (c) Levent Erkok
+-- License   : BSD3
+-- Maintainer: erkokl@gmail.com
+-- Stability : experimental
 --
 -- Internal data-structures for the sbv library
 -----------------------------------------------------------------------------
 
 {-# LANGUAGE CPP                   #-}
-{-# LANGUAGE TypeSynonymInstances  #-}
-{-# LANGUAGE TypeFamilies          #-}
-{-# LANGUAGE TypeOperators         #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
+{-# LANGUAGE DefaultSignatures     #-}
+{-# LANGUAGE DeriveAnyClass        #-}
+{-# LANGUAGE DeriveGeneric         #-}
 {-# LANGUAGE FlexibleContexts      #-}
 {-# LANGUAGE FlexibleInstances     #-}
 {-# LANGUAGE InstanceSigs          #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE PatternGuards         #-}
-{-# LANGUAGE DefaultSignatures     #-}
-{-# LANGUAGE NamedFieldPuns        #-}
-{-# LANGUAGE DeriveAnyClass        #-}
-{-# LANGUAGE DeriveGeneric         #-}
+{-# LANGUAGE ScopedTypeVariables   #-}
+{-# LANGUAGE TypeApplications      #-}
+{-# LANGUAGE TypeFamilies          #-}
+{-# LANGUAGE TypeOperators         #-}
+
+{-# OPTIONS_GHC -Wall -Werror #-}
 
 module Data.SBV.Core.Data
  ( SBool, SWord8, SWord16, SWord32, SWord64
  , SInt8, SInt16, SInt32, SInt64, SInteger, SReal, SFloat, SDouble, SChar, SString, SList
+ , SEither, SMaybe
+ , STuple, STuple2, STuple3, STuple4, STuple5, STuple6, STuple7, STuple8
+ , RCSet(..), SSet
  , nan, infinity, sNaN, sInfinity, RoundingMode(..), SRoundingMode
  , sRoundNearestTiesToEven, sRoundNearestTiesToAway, sRoundTowardPositive, sRoundTowardNegative, sRoundTowardZero
  , sRNE, sRNA, sRTP, sRTN, sRTZ
- , SymWord(..)
- , CW(..), CWVal(..), AlgReal(..), AlgRealPoly, ExtCW(..), GeneralizedCW(..), isRegularCW, cwSameType, cwToBool
- , mkConstCW ,liftCW2, mapCW, mapCW2
- , SW(..), trueSW, falseSW, trueCW, falseCW, normCW
+ , SymVal(..)
+ , CV(..), CVal(..), AlgReal(..), AlgRealPoly(..), ExtCV(..), GeneralizedCV(..), isRegularCV, cvSameType, cvToBool
+ , mkConstCV ,liftCV2, mapCV, mapCV2
+ , SV(..), trueSV, falseSV, trueCV, falseCV, normCV
  , SVal(..)
+ , sTrue, sFalse, sNot, (.&&), (.||), (.<+>), (.~&), (.~|), (.=>), (.<=>), sAnd, sOr, sAny, sAll, fromBool
  , SBV(..), NodeId(..), mkSymSBV
  , ArrayContext(..), ArrayInfo, SymArray(..), SFunArray(..), SArray(..)
- , sbvToSW, sbvToSymSW, forceSWArg
+ , sbvToSV, sbvToSymSV, forceSVArg
  , SBVExpr(..), newExpr
  , cache, Cached, uncache, uncacheAI, HasKind(..)
  , Op(..), PBOp(..), FPOp(..), StrOp(..), SeqOp(..), RegExp(..), NamedSymVar, getTableIndex
@@ -51,20 +56,21 @@ module Data.SBV.Core.Data
  , extractSymbolicSimulationState
  , SMTScript(..), Solver(..), SMTSolver(..), SMTResult(..), SMTModel(..), SMTConfig(..)
  , OptimizeStyle(..), Penalty(..), Objective(..)
- , QueryState(..), Query(..), SMTProblem(..)
+ , QueryState(..), QueryT(..), SMTProblem(..)
  ) where
 
 import GHC.Generics (Generic)
 import GHC.Exts     (IsList(..))
 
-import Control.DeepSeq      (NFData(..))
-import Control.Monad.Reader (ask)
-import Control.Monad.Trans  (liftIO)
-import Data.Int             (Int8, Int16, Int32, Int64)
-import Data.Word            (Word8, Word16, Word32, Word64)
-import Data.List            (elemIndex)
-import Data.Maybe           (fromMaybe)
-import Data.Typeable        (Typeable)
+import Control.DeepSeq        (NFData(..))
+import Control.Monad.Trans    (liftIO)
+import Data.Int               (Int8, Int16, Int32, Int64)
+import Data.Word              (Word8, Word16, Word32, Word64)
+import Data.List              (elemIndex)
+import Data.Maybe             (fromMaybe)
+
+import Data.Proxy
+import Data.Typeable          (Typeable)
 
 import qualified Data.Generics as G    (Data(..))
 
@@ -81,7 +87,6 @@ import Data.SBV.Control.Types
 import Data.SBV.SMT.SMTLibNames
 
 import Data.SBV.Utils.Lib
-import Data.SBV.Utils.Boolean
 
 -- | Get the current path condition
 getPathCondition :: State -> SBool
@@ -158,8 +163,43 @@ type SString = SBV String
 -- Note that lists can be nested, i.e., we do allow lists of lists of ... items.
 type SList a = SBV [a]
 
+-- | Symbolic 'Either'
+type SEither a b = SBV (Either a b)
+
+-- | Symbolic 'Maybe'
+type SMaybe a = SBV (Maybe a)
+
+-- | Symbolic 'Data.Set'. Note that we use 'RCSet', which supports
+-- both regular sets and complements, i.e., those obtained from the
+-- universal set (of the right type) by removing elements.
+type SSet a = SBV (RCSet a)
+
+-- | Symbolic 2-tuple. NB. 'STuple' and 'STuple2' are equivalent.
+type STuple a b = SBV (a, b)
+
+-- | Symbolic 2-tuple. NB. 'STuple' and 'STuple2' are equivalent.
+type STuple2 a b = SBV (a, b)
+
+-- | Symbolic 3-tuple.
+type STuple3 a b c = SBV (a, b, c)
+
+-- | Symbolic 4-tuple.
+type STuple4 a b c d = SBV (a, b, c, d)
+
+-- | Symbolic 5-tuple.
+type STuple5 a b c d e = SBV (a, b, c, d, e)
+
+-- | Symbolic 6-tuple.
+type STuple6 a b c d e f = SBV (a, b, c, d, e, f)
+
+-- | Symbolic 7-tuple.
+type STuple7 a b c d e f g = SBV (a, b, c, d, e, f, g)
+
+-- | Symbolic 8-tuple.
+type STuple8 a b c d e f g h = SBV (a, b, c, d, e, f, g, h)
+
 -- | IsList instance allows list literals to be written compactly.
-instance SymWord [a] => IsList (SList a) where
+instance SymVal [a] => IsList (SList a) where
   type Item (SList a) = a
   fromList = literal
   toList x = fromMaybe (error "IsList.toList used in a symbolic context!") (unliteral x)
@@ -176,28 +216,90 @@ infinity = 1/0
 
 -- | Symbolic variant of Not-A-Number. This value will inhabit both
 -- 'SDouble' and 'SFloat'.
-sNaN :: (Floating a, SymWord a) => SBV a
+sNaN :: (Floating a, SymVal a) => SBV a
 sNaN = literal nan
 
 -- | Symbolic variant of infinity. This value will inhabit both
 -- 'SDouble' and 'SFloat'.
-sInfinity :: (Floating a, SymWord a) => SBV a
+sInfinity :: (Floating a, SymVal a) => SBV a
 sInfinity = literal infinity
 
 -- | Internal representation of a symbolic simulation result
 newtype SMTProblem = SMTProblem {smtLibPgm :: SMTConfig -> SMTLibPgm} -- ^ SMTLib representation, given the config
 
--- Boolean combinators
-instance Boolean SBool where
-  true  = SBV (svBool True)
-  false = SBV (svBool False)
-  bnot (SBV b) = SBV (svNot b)
-  SBV a &&& SBV b = SBV (svAnd a b)
-  SBV a ||| SBV b = SBV (svOr a b)
-  SBV a <+> SBV b = SBV (svXOr a b)
+-- | Symbolic 'True'
+sTrue :: SBool
+sTrue = SBV (svBool True)
+
+-- | Symbolic 'False'
+sFalse :: SBool
+sFalse = SBV (svBool False)
+
+-- | Symbolic boolean negation
+sNot :: SBool -> SBool
+sNot (SBV b) = SBV (svNot b)
+
+-- | Symbolic conjunction
+infixr 3 .&&
+(.&&) :: SBool -> SBool -> SBool
+SBV x .&& SBV y = SBV (x `svAnd` y)
+
+-- | Symbolic disjunction
+infixr 2 .||
+(.||) :: SBool -> SBool -> SBool
+SBV x .|| SBV y = SBV (x `svOr` y)
+
+-- | Symbolic logical xor
+infixl 6 .<+>
+(.<+>) :: SBool -> SBool -> SBool
+SBV x .<+> SBV y = SBV (x `svXOr` y)
+
+-- | Symbolic nand
+infixr 3 .~&
+(.~&) :: SBool -> SBool -> SBool
+x .~& y = sNot (x .&& y)
+
+-- | Symbolic nor
+infixr 2 .~|
+(.~|) :: SBool -> SBool -> SBool
+x .~| y = sNot (x .|| y)
+
+-- | Symbolic implication
+infixr 1 .=>
+(.=>) :: SBool -> SBool -> SBool
+x .=> y = sNot x .|| y
+-- NB. Do *not* try to optimize @x .=> x = True@ here! If constants go through, it'll get simplified.
+-- The case "x .=> x" can hit is extremely rare, and the getAllSatResult function relies on this
+-- trick to generate constraints in the unlucky case of ui-function models.
+
+-- | Symbolic boolean equivalence
+infixr 1 .<=>
+(.<=>) :: SBool -> SBool -> SBool
+x .<=> y = (x .&& y) .|| (sNot x .&& sNot y)
+
+-- | Conversion from 'Bool' to 'SBool'
+fromBool :: Bool -> SBool
+fromBool True  = sTrue
+fromBool False = sFalse
+
+-- | Generalization of 'and'
+sAnd :: [SBool] -> SBool
+sAnd = foldr (.&&) sTrue
+
+-- | Generalization of 'or'
+sOr :: [SBool] -> SBool
+sOr  = foldr (.||) sFalse
+
+-- | Generalization of 'any'
+sAny :: (a -> SBool) -> [a] -> SBool
+sAny f = sOr  . map f
+
+-- | Generalization of 'all'
+sAll :: (a -> SBool) -> [a] -> SBool
+sAll f = sAnd . map f
 
 -- | 'RoundingMode' can be used symbolically
-instance SymWord RoundingMode
+instance SymVal RoundingMode
 
 -- | The symbolic variant of 'RoundingMode'
 type SRoundingMode = SBV RoundingMode
@@ -255,39 +357,39 @@ instance Eq (SBV a) where
   SBV a == SBV b = a == b
   SBV a /= SBV b = a /= b
 
-instance HasKind (SBV a) where
-  kindOf (SBV (SVal k _)) = k
+instance HasKind a => HasKind (SBV a) where
+  kindOf _ = kindOf (Proxy @a)
 
 -- | Convert a symbolic value to a symbolic-word
-sbvToSW :: State -> SBV a -> IO SW
-sbvToSW st (SBV s) = svToSW st s
+sbvToSV :: State -> SBV a -> IO SV
+sbvToSV st (SBV s) = svToSV st s
 
 -------------------------------------------------------------------------
 -- * Symbolic Computations
 -------------------------------------------------------------------------
 
--- | Create a symbolic variable.
-mkSymSBV :: forall a. Maybe Quantifier -> Kind -> Maybe String -> Symbolic (SBV a)
-mkSymSBV mbQ k mbNm = SBV <$> (ask >>= liftIO . svMkSymVar mbQ k mbNm)
+-- | Generalization of 'Data.SBV.mkSymSBV'
+mkSymSBV :: forall a m. MonadSymbolic m => Maybe Quantifier -> Kind -> Maybe String -> m (SBV a)
+mkSymSBV mbQ k mbNm = SBV <$> (symbolicEnv >>= liftIO . svMkSymVar mbQ k mbNm)
 
--- | Convert a symbolic value to an SW, inside the Symbolic monad
-sbvToSymSW :: SBV a -> Symbolic SW
-sbvToSymSW sbv = do
-        st <- ask
-        liftIO $ sbvToSW st sbv
+-- | Generalization of 'Data.SBV.sbvToSymSW'
+sbvToSymSV :: MonadSymbolic m => SBV a -> m SV
+sbvToSymSV sbv = do
+        st <- symbolicEnv
+        liftIO $ sbvToSV st sbv
 
 -- | Actions we can do in a context: Either at problem description
 -- time or while we are dynamically querying. 'Symbolic' and 'Query' are
 -- two instances of this class. Note that we use this mechanism
 -- internally and do not export it from SBV.
 class SolverContext m where
-   -- | Add a constraint, any satisfying instance must satisfy this condition
-   constrain       :: SBool -> m ()
-   -- | Add a soft constraint. The solver will try to satisfy this condition if possible, but won't if it cannot
-   softConstrain   :: SBool -> m ()
+   -- | Add a constraint, any satisfying instance must satisfy this condition.
+   constrain :: SBool -> m ()
+   -- | Add a soft constraint. The solver will try to satisfy this condition if possible, but won't if it cannot.
+   softConstrain :: SBool -> m ()
    -- | Add a named constraint. The name is used in unsat-core extraction.
    namedConstraint :: String -> SBool -> m ()
-   -- | Add a constraint, with arbitrary attributes. Used in interpolant generation.
+   -- | Add a constraint, with arbitrary attributes.
    constrainWithAttribute :: [(String, String)] -> SBool -> m ()
    -- | Set info. Example: @setInfo ":status" ["unsat"]@.
    setInfo :: String -> [String] -> m ()
@@ -301,6 +403,10 @@ class SolverContext m where
    -- is in milliseconds. Also see the function 'Data.SBV.Control.timeOut' for finer level
    -- control of time-outs, directly from SBV.
    setTimeOut :: Integer -> m ()
+   -- | Get the state associated with this context
+   contextState :: m State
+
+   {-# MINIMAL constrain, softConstrain, namedConstraint, constrainWithAttribute, setOption, contextState #-}
 
    -- time-out, logic, and info are  simply options in our implementation, so default implementation suffices
    setTimeOut t = setOption $ OptionKeyword ":timeout" [show t]
@@ -309,9 +415,8 @@ class SolverContext m where
 
 -- | A class representing what can be returned from a symbolic computation.
 class Outputtable a where
-  -- | Mark an interim result as an output. Useful when constructing Symbolic programs
-  -- that return multiple values, or when the result is programmatically computed.
-  output :: a -> Symbolic a
+  -- | Generalization of 'Data.SBV.output'
+  output :: MonadSymbolic m => a -> m a
 
 instance Outputtable (SBV a) where
   output i = do
@@ -346,90 +451,106 @@ instance (Outputtable a, Outputtable b, Outputtable c, Outputtable d, Outputtabl
   output = mlift8 (,,,,,,,) output output output output output output output output
 
 -------------------------------------------------------------------------------
--- * Symbolic Words
+-- * Symbolic Values
 -------------------------------------------------------------------------------
--- | A 'SymWord' is a potential symbolic bitvector that can be created instances of
--- to be fed to a symbolic program. Note that these methods are typically not needed
--- in casual uses with 'Data.SBV.prove', 'Data.SBV.sat', 'Data.SBV.allSat' etc, as
--- default instances automatically provide the necessary bits.
-class (HasKind a, Ord a, Typeable a) => SymWord a where
-  -- | Create a user named input (universal)
-  forall :: String -> Symbolic (SBV a)
-  -- | Create an automatically named input
-  forall_ :: Symbolic (SBV a)
-  -- | Get a bunch of new words
-  mkForallVars :: Int -> Symbolic [SBV a]
-  -- | Create an existential variable
-  exists  :: String -> Symbolic (SBV a)
-  -- | Create an automatically named existential variable
-  exists_ :: Symbolic (SBV a)
-  -- | Create a bunch of existentials
-  mkExistVars :: Int -> Symbolic [SBV a]
-  -- | Create a free variable, universal in a proof, existential in sat
-  free :: String -> Symbolic (SBV a)
-  -- | Create an unnamed free variable, universal in proof, existential in sat
-  free_ :: Symbolic (SBV a)
-  -- | Create a bunch of free vars
-  mkFreeVars :: Int -> Symbolic [SBV a]
-  -- | Similar to free; Just a more convenient name
-  symbolic  :: String -> Symbolic (SBV a)
-  -- | Similar to mkFreeVars; but automatically gives names based on the strings
-  symbolics :: [String] -> Symbolic [SBV a]
+-- | A 'SymVal' is a potential symbolic value that can be created instances of to be fed to a symbolic program.
+class (HasKind a, Typeable a) => SymVal a where
+  -- | Generalization of 'Data.SBV.mkSymVal'
+  mkSymVal :: MonadSymbolic m => Maybe Quantifier -> Maybe String -> m (SBV a)
   -- | Turn a literal constant to symbolic
   literal :: a -> SBV a
-  -- | Extract a literal, if the value is concrete
-  unliteral :: SBV a -> Maybe a
-  -- | Extract a literal, from a CW representation
-  fromCW :: CW -> a
-  -- | Is the symbolic word concrete?
-  isConcrete :: SBV a -> Bool
-  -- | Is the symbolic word really symbolic?
-  isSymbolic :: SBV a -> Bool
+  -- | Extract a literal, from a CV representation
+  fromCV :: CV -> a
   -- | Does it concretely satisfy the given predicate?
   isConcretely :: SBV a -> (a -> Bool) -> Bool
-  -- | One stop allocator
-  mkSymWord :: Maybe Quantifier -> Maybe String -> Symbolic (SBV a)
 
-  -- minimal complete definition:: Nothing.
-  -- Giving no instances is ok when defining an uninterpreted/enumerated sort, but otherwise you really
-  -- want to define: literal, fromCW, mkSymWord
-  forall   = mkSymWord (Just ALL) . Just
-  forall_  = mkSymWord (Just ALL)   Nothing
-  exists   = mkSymWord (Just EX)  . Just
-  exists_  = mkSymWord (Just EX)    Nothing
-  free     = mkSymWord Nothing    . Just
-  free_    = mkSymWord Nothing      Nothing
-  mkForallVars n = mapM (const forall_) [1 .. n]
-  mkExistVars n  = mapM (const exists_) [1 .. n]
-  mkFreeVars n   = mapM (const free_)   [1 .. n]
-  symbolic       = free
-  symbolics      = mapM symbolic
-  unliteral (SBV (SVal _ (Left c)))  = Just $ fromCW c
-  unliteral _                        = Nothing
-  isConcrete (SBV (SVal _ (Left _))) = True
-  isConcrete _                       = False
-  isSymbolic = not . isConcrete
+  -- minimal complete definition: Nothing.
+  -- Giving no instances is okay when defining an uninterpreted/enumerated sort, but otherwise you really
+  -- want to define: literal, fromCV, mkSymVal
+
+  default mkSymVal :: (MonadSymbolic m, Read a, G.Data a) => Maybe Quantifier -> Maybe String -> m (SBV a)
+  mkSymVal mbQ mbNm = SBV <$> (symbolicEnv >>= liftIO . svMkSymVar mbQ k mbNm)
+    where -- NB.A call of the form
+          --      constructUKind (Proxy @a)
+          -- would be wrong here, as it would uninterpret the Proxy datatype!
+          -- So, we have to use the dreaded undefined value in this case.
+          k = constructUKind (undefined :: a)
+
+  default literal :: Show a => a -> SBV a
+  literal x = let k@(KUninterpreted  _ conts) = kindOf x
+                  sx                          = show x
+                  mbIdx = case conts of
+                            Right xs -> sx `elemIndex` xs
+                            _        -> Nothing
+              in SBV $ SVal k (Left (CV k (CUserSort (mbIdx, sx))))
+
+  default fromCV :: Read a => CV -> a
+  fromCV (CV _ (CUserSort (_, s))) = read s
+  fromCV cv                        = error $ "Cannot convert CV " ++ show cv ++ " to kind " ++ show (kindOf (Proxy @a))
+
   isConcretely s p
     | Just i <- unliteral s = p i
     | True                  = False
 
-  default literal :: Show a => a -> SBV a
-  literal x = let k@(KUserSort  _ conts) = kindOf x
-                  sx                     = show x
-                  mbIdx = case conts of
-                            Right xs -> sx `elemIndex` xs
-                            _        -> Nothing
-              in SBV $ SVal k (Left (CW k (CWUserSort (mbIdx, sx))))
+  -- | Generalization of 'Data.SBV.forall'
+  forall :: MonadSymbolic m => String -> m (SBV a)
+  forall = mkSymVal (Just ALL) . Just
 
-  default fromCW :: Read a => CW -> a
-  fromCW (CW _ (CWUserSort (_, s))) = read s
-  fromCW cw                         = error $ "Cannot convert CW " ++ show cw ++ " to kind " ++ show (kindOf (undefined :: a))
+  -- | Generalization of 'Data.SBV.forall_'
+  forall_ :: MonadSymbolic m => m (SBV a)
+  forall_ = mkSymVal (Just ALL) Nothing
 
-  default mkSymWord :: (Read a, G.Data a) => Maybe Quantifier -> Maybe String -> Symbolic (SBV a)
-  mkSymWord mbQ mbNm = SBV <$> (ask >>= liftIO . svMkSymVar mbQ k mbNm)
-    where k = constructUKind (undefined :: a)
+  -- | Generalization of 'Data.SBV.mkForallVars'
+  mkForallVars :: MonadSymbolic m => Int -> m [SBV a]
+  mkForallVars n = mapM (const forall_) [1 .. n]
 
-instance (Random a, SymWord a) => Random (SBV a) where
+  -- | Generalization of 'Data.SBV.exists'
+  exists :: MonadSymbolic m => String -> m (SBV a)
+  exists = mkSymVal (Just EX) . Just
+
+  -- | Generalization of 'Data.SBV.exists_'
+  exists_ :: MonadSymbolic m => m (SBV a)
+  exists_ = mkSymVal (Just EX) Nothing
+
+  -- | Generalization of 'Data.SBV.mkExistVars'
+  mkExistVars :: MonadSymbolic m => Int -> m [SBV a]
+  mkExistVars n = mapM (const exists_) [1 .. n]
+
+  -- | Generalization of 'Data.SBV.free'
+  free :: MonadSymbolic m => String -> m (SBV a)
+  free = mkSymVal Nothing . Just
+
+  -- | Generalization of 'Data.SBV.free_'
+  free_ :: MonadSymbolic m => m (SBV a)
+  free_ = mkSymVal Nothing Nothing
+
+  -- | Generalization of 'Data.SBV.mkFreeVars'
+  mkFreeVars :: MonadSymbolic m => Int -> m [SBV a]
+  mkFreeVars n = mapM (const free_) [1 .. n]
+
+  -- | Generalization of 'Data.SBV.symbolic'
+  symbolic :: MonadSymbolic m => String -> m (SBV a)
+  symbolic = free
+
+  -- | Generalization of 'Data.SBV.symbolics'
+  symbolics :: MonadSymbolic m => [String] -> m [SBV a]
+  symbolics = mapM symbolic
+
+  -- | Extract a literal, if the value is concrete
+  unliteral :: SBV a -> Maybe a
+  unliteral (SBV (SVal _ (Left c))) = Just $ fromCV c
+  unliteral _                       = Nothing
+
+  -- | Is the symbolic word concrete?
+  isConcrete :: SBV a -> Bool
+  isConcrete (SBV (SVal _ (Left _))) = True
+  isConcrete _                       = False
+
+  -- | Is the symbolic word really symbolic?
+  isSymbolic :: SBV a -> Bool
+  isSymbolic = not . isConcrete
+
+instance (Random a, SymVal a) => Random (SBV a) where
   randomR (l, h) g = case (unliteral l, unliteral h) of
                        (Just lb, Just hb) -> let (v, g') = randomR (lb, hb) g in (literal (v :: a), g')
                        _                  -> error "SBV.Random: Cannot generate random values with symbolic bounds"
@@ -477,30 +598,34 @@ instance (Random a, SymWord a) => Random (SBV a) where
 --      'SFunArray', SBV only generates code for individual elements and the array itself never
 --      shows up in the resulting SMTLib program. This puts more onus on the SBV side and might
 --      have some performance impacts, but it might generate problems that are easier for the SMT
---      solvers to handle. 
+--      solvers to handle.
 --
 -- As a rule of thumb, try 'SArray' first. These should generate compact code. However, if
 -- the backend solver has hard time solving the generated problems, switch to
 -- 'SFunArray'. If you still have issues, please report so we can see what the problem might be!
 class SymArray array where
-  -- | Create a new anonymous array, possibly with a default initial value.
-  newArray_      :: (HasKind a, HasKind b) => Maybe (SBV b) -> Symbolic (array a b)
-  -- | Create a named new array, possibly with a default initial value.
-  newArray       :: (HasKind a, HasKind b) => String -> Maybe (SBV b) -> Symbolic (array a b)
+  -- | Generalization of 'Data.SBV.newArray_'
+  newArray_      :: (MonadSymbolic m, HasKind a, HasKind b) => Maybe (SBV b) -> m (array a b)
+  -- | Generalization of 'Data.SBV.newArray'
+  newArray       :: (MonadSymbolic m, HasKind a, HasKind b) => String -> Maybe (SBV b) -> m (array a b)
   -- | Read the array element at @a@
   readArray      :: array a b -> SBV a -> SBV b
   -- | Update the element at @a@ to be @b@
-  writeArray     :: SymWord b => array a b -> SBV a -> SBV b -> array a b
+  writeArray     :: SymVal b => array a b -> SBV a -> SBV b -> array a b
   -- | Merge two given arrays on the symbolic condition
   -- Intuitively: @mergeArrays cond a b = if cond then a else b@.
   -- Merging pushes the if-then-else choice down on to elements
-  mergeArrays    :: SymWord b => SBV Bool -> array a b -> array a b -> array a b
+  mergeArrays    :: SymVal b => SBV Bool -> array a b -> array a b -> array a b
   -- | Internal function, not exported to the user
   newArrayInState :: (HasKind a, HasKind b) => Maybe String -> Maybe (SBV b) -> State -> IO (array a b)
 
-  {-# MINIMAL readArray, writeArray, mergeArrays, newArrayInState #-}
-  newArray_   mbVal = ask >>= liftIO . newArrayInState Nothing   mbVal
-  newArray nm mbVal = ask >>= liftIO . newArrayInState (Just nm) mbVal
+  {-# MINIMAL readArray, writeArray, mergeArrays, ((newArray_, newArray) | newArrayInState) #-}
+  newArray_   mbVal = symbolicEnv >>= liftIO . newArrayInState Nothing   mbVal
+  newArray nm mbVal = symbolicEnv >>= liftIO . newArrayInState (Just nm) mbVal
+
+  -- Despite our MINIMAL pragma and default implementations for newArray_ and
+  -- newArray, we must provide a dummy implementation for newArrayInState:
+  newArrayInState = error "undefined: newArrayInState"
 
 -- | Arrays implemented in terms of SMT-arrays: <http://smtlib.cs.uiowa.edu/theories-ArraysEx.shtml>
 --
@@ -519,7 +644,7 @@ class SymArray array where
 newtype SArray a b = SArray { unSArray :: SArr }
 
 instance (HasKind a, HasKind b) => Show (SArray a b) where
-  show SArray{} = "SArray<" ++ showType (undefined :: a) ++ ":" ++ showType (undefined :: b) ++ ">"
+  show SArray{} = "SArray<" ++ showType (Proxy @a) ++ ":" ++ showType (Proxy @b) ++ ">"
 
 instance SymArray SArray where
   readArray   (SArray arr) (SBV a)               = SBV (readSArr arr a)
@@ -531,8 +656,8 @@ instance SymArray SArray where
                                      SArray <$> newSArr st (aknd, bknd) (mkNm mbNm) (unSBV <$> mbVal)
      where mkNm Nothing   t = "array_" ++ show t
            mkNm (Just nm) _ = nm
-           aknd = kindOf (undefined :: a)
-           bknd = kindOf (undefined :: b)
+           aknd = kindOf (Proxy @a)
+           bknd = kindOf (Proxy @b)
 
 -- | Arrays implemented internally, without translating to SMT-Lib functions:
 --
@@ -552,7 +677,7 @@ instance SymArray SArray where
 newtype SFunArray a b = SFunArray { unSFunArray :: SFunArr }
 
 instance (HasKind a, HasKind b) => Show (SFunArray a b) where
-  show SFunArray{} = "SFunArray<" ++ showType (undefined :: a) ++ ":" ++ showType (undefined :: b) ++ ">"
+  show SFunArray{} = "SFunArray<" ++ showType (Proxy @a) ++ ":" ++ showType (Proxy @b) ++ ">"
 
 instance SymArray SFunArray where
   readArray   (SFunArray arr) (SBV a)             = SBV (readSFunArr arr a)
@@ -564,5 +689,5 @@ instance SymArray SFunArray where
                                      SFunArray <$> newSFunArr st (aknd, bknd) (mkNm mbNm) (unSBV <$> mbVal)
     where mkNm Nothing t   = "funArray_" ++ show t
           mkNm (Just nm) _ = nm
-          aknd = kindOf (undefined :: a)
-          bknd = kindOf (undefined :: b)
+          aknd = kindOf (Proxy @a)
+          bknd = kindOf (Proxy @b)
