@@ -1,19 +1,21 @@
 -----------------------------------------------------------------------------
 -- |
--- Module      :  Data.SBV.SMT.SMT
--- Copyright   :  (c) Levent Erkok
--- License     :  BSD3
--- Maintainer  :  erkokl@gmail.com
--- Stability   :  experimental
+-- Module    : Data.SBV.SMT.SMT
+-- Copyright : (c) Levent Erkok
+-- License   : BSD3
+-- Maintainer: erkokl@gmail.com
+-- Stability : experimental
 --
 -- Abstraction of SMT solvers
 -----------------------------------------------------------------------------
 
-{-# LANGUAGE ScopedTypeVariables        #-}
 {-# LANGUAGE DefaultSignatures          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE Rank2Types                 #-}
 {-# LANGUAGE NamedFieldPuns             #-}
+{-# LANGUAGE Rank2Types                 #-}
+{-# LANGUAGE ScopedTypeVariables        #-}
+
+{-# OPTIONS_GHC -Wall -Werror #-}
 
 module Data.SBV.SMT.SMT (
        -- * Model extraction
@@ -21,7 +23,7 @@ module Data.SBV.SMT.SMT (
        , SatModel(..), genParse
        , extractModels, getModelValues
        , getModelDictionaries, getModelUninterpretedValues
-       , displayModels, showModel
+       , displayModels, showModel, shCV, showModelDictionary
 
        -- * Standard prover engine
        , standardEngine
@@ -43,7 +45,7 @@ import Control.Monad      (zipWithM)
 import Data.Char          (isSpace)
 import Data.Maybe         (fromMaybe, isJust)
 import Data.Int           (Int8, Int16, Int32, Int64)
-import Data.List          (intercalate, isPrefixOf)
+import Data.List          (intercalate, isPrefixOf, transpose)
 import Data.Word          (Word8, Word16, Word32, Word64)
 
 import Data.IORef (readIORef, writeIORef)
@@ -61,6 +63,8 @@ import qualified Data.Map.Strict as M
 import Data.SBV.Core.AlgReals
 import Data.SBV.Core.Data
 import Data.SBV.Core.Symbolic (SMTEngine, State(..))
+import Data.SBV.Core.Concrete (showCV)
+import Data.SBV.Core.Kind     (showBaseKind)
 
 import Data.SBV.SMT.Utils     (showTimeoutValue, alignPlain, debug, mergeSExpr, SBVException(..))
 
@@ -73,11 +77,11 @@ import qualified System.Timeout as Timeout (timeout)
 
 -- | Extract the final configuration from a result
 resultConfig :: SMTResult -> SMTConfig
-resultConfig (Unsatisfiable c _) = c
-resultConfig (Satisfiable   c _) = c
-resultConfig (SatExtField   c _) = c
-resultConfig (Unknown       c _) = c
-resultConfig (ProofError    c _) = c
+resultConfig (Unsatisfiable c _  ) = c
+resultConfig (Satisfiable   c _  ) = c
+resultConfig (SatExtField   c _  ) = c
+resultConfig (Unknown       c _  ) = c
+resultConfig (ProofError    c _ _) = c
 
 -- | A 'Data.SBV.prove' call results in a 'ThmResult'
 newtype ThmResult = ThmResult SMTResult
@@ -90,8 +94,9 @@ newtype SatResult = SatResult SMTResult
 
 -- | An 'Data.SBV.allSat' call results in a 'AllSatResult'. The first boolean says whether we
 -- hit the max-model limit as we searched. The second boolean says whether
--- there were prefix-existentials.
-newtype AllSatResult = AllSatResult (Bool, Bool, [SMTResult])
+-- there were prefix-existentials. The third boolean says whether we stopped because
+-- the solver returned 'Unknown'.
+newtype AllSatResult = AllSatResult (Bool, Bool, Bool, [SMTResult])
 
 -- | A 'Data.SBV.safe' call results in a 'SafeResult'
 newtype SafeResult   = SafeResult   (Maybe String, String, SMTResult)
@@ -126,17 +131,22 @@ instance Show SafeResult where
 -- The Show instance of AllSatResults. Note that we have to be careful in being lazy enough
 -- as the typical use case is to pull results out as they become available.
 instance Show AllSatResult where
-  show (AllSatResult (l, e, xs)) = go (0::Int) xs
-    where uniqueWarn | e    = " (Unique up to prefix existentials.)"
-                     | True = ""
+  show (AllSatResult (l, e, u, xs)) = go (0::Int) xs
+    where warnings = case (e, u) of
+                       (False, False) -> ""
+                       (False, True)  -> " (Search stopped since solver has returned unknown.)"
+                       (True,  False) -> " (Unique up to prefix existentials.)"
+                       (True,  True)  -> " (Search stopped becase solver has returned unknown, only prefix existentials were considered.)"
+
           go c (s:ss) = let c'      = c+1
                             (ok, o) = sh c' s
                         in c' `seq` if ok then o ++ "\n" ++ go c' ss else o
           go c []     = case (l, c) of
-                          (True,  _) -> "Search stopped since model count request was reached." ++ uniqueWarn
+                          (True,  _) -> "Search stopped since model count request was reached." ++ warnings
                           (False, 0) -> "No solutions found."
-                          (False, 1) -> "This is the only solution." ++ uniqueWarn
-                          (False, _) -> "Found " ++ show c ++ " different solutions." ++ uniqueWarn
+                          (False, 1) -> "This is the only solution." ++ warnings
+                          (False, _) -> "Found " ++ show c ++ " different solutions." ++ warnings
+
           sh i c = (ok, showSMTResult "Unsatisfiable"
                                       "Unknown"
                                       ("Solution #" ++ show i ++ ":\nSatisfiable") ("Solution #" ++ show i ++ ":\n")
@@ -153,7 +163,7 @@ instance Show OptimizeResult where
 
                IndependentResult   rs  -> multi "objectives" (map (uncurry shI) rs)
 
-               ParetoResult (False, [r]) -> sh (\s -> "Unique pareto front: " ++ s) r
+               ParetoResult (False, [r]) -> sh ("Unique pareto front: " ++) r
                ParetoResult (False, rs)  -> multi "pareto optimal values" (zipWith shP [(1::Int)..] rs)
                ParetoResult (True,  rs)  ->    multi "pareto optimal values" (zipWith shP [(1::Int)..] rs)
                                            ++ "\n*** Note: Pareto-front extraction was terminated as requested by the user."
@@ -172,94 +182,94 @@ instance Show OptimizeResult where
                                     (tag "Optimal in an extension field:" ++ "\n")
 
 -- | Instances of 'SatModel' can be automatically extracted from models returned by the
--- solvers. The idea is that the sbv infrastructure provides a stream of CW's (constant-words)
+-- solvers. The idea is that the sbv infrastructure provides a stream of CV's (constant values)
 -- coming from the solver, and the type @a@ is interpreted based on these constants. Many typical
 -- instances are already provided, so new instances can be declared with relative ease.
 --
--- Minimum complete definition: 'parseCWs'
+-- Minimum complete definition: 'parseCVs'
 class SatModel a where
   -- | Given a sequence of constant-words, extract one instance of the type @a@, returning
   -- the remaining elements untouched. If the next element is not what's expected for this
   -- type you should return 'Nothing'
-  parseCWs  :: [CW] -> Maybe (a, [CW])
+  parseCVs  :: [CV] -> Maybe (a, [CV])
   -- | Given a parsed model instance, transform it using @f@, and return the result.
   -- The default definition for this method should be sufficient in most use cases.
-  cvtModel  :: (a -> Maybe b) -> Maybe (a, [CW]) -> Maybe (b, [CW])
+  cvtModel  :: (a -> Maybe b) -> Maybe (a, [CV]) -> Maybe (b, [CV])
   cvtModel f x = x >>= \(a, r) -> f a >>= \b -> return (b, r)
 
-  default parseCWs :: Read a => [CW] -> Maybe (a, [CW])
-  parseCWs (CW _ (CWUserSort (_, s)) : r) = Just (read s, r)
-  parseCWs _                              = Nothing
+  default parseCVs :: Read a => [CV] -> Maybe (a, [CV])
+  parseCVs (CV _ (CUserSort (_, s)) : r) = Just (read s, r)
+  parseCVs _                             = Nothing
 
--- | Parse a signed/sized value from a sequence of CWs
-genParse :: Integral a => Kind -> [CW] -> Maybe (a, [CW])
-genParse k (x@(CW _ (CWInteger i)):r) | kindOf x == k = Just (fromIntegral i, r)
-genParse _ _                                          = Nothing
+-- | Parse a signed/sized value from a sequence of CVs
+genParse :: Integral a => Kind -> [CV] -> Maybe (a, [CV])
+genParse k (x@(CV _ (CInteger i)):r) | kindOf x == k = Just (fromIntegral i, r)
+genParse _ _                                         = Nothing
 
 -- | Base case for 'SatModel' at unit type. Comes in handy if there are no real variables.
 instance SatModel () where
-  parseCWs xs = return ((), xs)
+  parseCVs xs = return ((), xs)
 
 -- | 'Bool' as extracted from a model
 instance SatModel Bool where
-  parseCWs xs = do (x, r) <- genParse KBool xs
+  parseCVs xs = do (x, r) <- genParse KBool xs
                    return ((x :: Integer) /= 0, r)
 
 -- | 'Word8' as extracted from a model
 instance SatModel Word8 where
-  parseCWs = genParse (KBounded False 8)
+  parseCVs = genParse (KBounded False 8)
 
 -- | 'Int8' as extracted from a model
 instance SatModel Int8 where
-  parseCWs = genParse (KBounded True 8)
+  parseCVs = genParse (KBounded True 8)
 
 -- | 'Word16' as extracted from a model
 instance SatModel Word16 where
-  parseCWs = genParse (KBounded False 16)
+  parseCVs = genParse (KBounded False 16)
 
 -- | 'Int16' as extracted from a model
 instance SatModel Int16 where
-  parseCWs = genParse (KBounded True 16)
+  parseCVs = genParse (KBounded True 16)
 
 -- | 'Word32' as extracted from a model
 instance SatModel Word32 where
-  parseCWs = genParse (KBounded False 32)
+  parseCVs = genParse (KBounded False 32)
 
 -- | 'Int32' as extracted from a model
 instance SatModel Int32 where
-  parseCWs = genParse (KBounded True 32)
+  parseCVs = genParse (KBounded True 32)
 
 -- | 'Word64' as extracted from a model
 instance SatModel Word64 where
-  parseCWs = genParse (KBounded False 64)
+  parseCVs = genParse (KBounded False 64)
 
 -- | 'Int64' as extracted from a model
 instance SatModel Int64 where
-  parseCWs = genParse (KBounded True 64)
+  parseCVs = genParse (KBounded True 64)
 
 -- | 'Integer' as extracted from a model
 instance SatModel Integer where
-  parseCWs = genParse KUnbounded
+  parseCVs = genParse KUnbounded
 
 -- | 'AlgReal' as extracted from a model
 instance SatModel AlgReal where
-  parseCWs (CW KReal (CWAlgReal i) : r) = Just (i, r)
-  parseCWs _                            = Nothing
+  parseCVs (CV KReal (CAlgReal i) : r) = Just (i, r)
+  parseCVs _                           = Nothing
 
 -- | 'Float' as extracted from a model
 instance SatModel Float where
-  parseCWs (CW KFloat (CWFloat i) : r) = Just (i, r)
-  parseCWs _                           = Nothing
+  parseCVs (CV KFloat (CFloat i) : r) = Just (i, r)
+  parseCVs _                          = Nothing
 
 -- | 'Double' as extracted from a model
 instance SatModel Double where
-  parseCWs (CW KDouble (CWDouble i) : r) = Just (i, r)
-  parseCWs _                             = Nothing
+  parseCVs (CV KDouble (CDouble i) : r) = Just (i, r)
+  parseCVs _                            = Nothing
 
--- | @CW@ as extracted from a model; trivial definition
-instance SatModel CW where
-  parseCWs (cw : r) = Just (cw, r)
-  parseCWs []       = Nothing
+-- | @CV@ as extracted from a model; trivial definition
+instance SatModel CV where
+  parseCVs (cv : r) = Just (cv, r)
+  parseCVs []       = Nothing
 
 -- | A rounding mode, extracted from a model. (Default definition suffices)
 instance SatModel RoundingMode
@@ -268,47 +278,47 @@ instance SatModel RoundingMode
 -- go as long as we can (maximal-munch). Note that this never fails, as
 -- we can always return the empty list!
 instance SatModel a => SatModel [a] where
-  parseCWs [] = Just ([], [])
-  parseCWs xs = case parseCWs xs of
-                  Just (a, ys) -> case parseCWs ys of
+  parseCVs [] = Just ([], [])
+  parseCVs xs = case parseCVs xs of
+                  Just (a, ys) -> case parseCVs ys of
                                     Just (as, zs) -> Just (a:as, zs)
                                     Nothing       -> Just ([], ys)
                   Nothing     -> Just ([], xs)
 
 -- | Tuples extracted from a model
 instance (SatModel a, SatModel b) => SatModel (a, b) where
-  parseCWs as = do (a, bs) <- parseCWs as
-                   (b, cs) <- parseCWs bs
+  parseCVs as = do (a, bs) <- parseCVs as
+                   (b, cs) <- parseCVs bs
                    return ((a, b), cs)
 
 -- | 3-Tuples extracted from a model
 instance (SatModel a, SatModel b, SatModel c) => SatModel (a, b, c) where
-  parseCWs as = do (a,      bs) <- parseCWs as
-                   ((b, c), ds) <- parseCWs bs
+  parseCVs as = do (a,      bs) <- parseCVs as
+                   ((b, c), ds) <- parseCVs bs
                    return ((a, b, c), ds)
 
 -- | 4-Tuples extracted from a model
 instance (SatModel a, SatModel b, SatModel c, SatModel d) => SatModel (a, b, c, d) where
-  parseCWs as = do (a,         bs) <- parseCWs as
-                   ((b, c, d), es) <- parseCWs bs
+  parseCVs as = do (a,         bs) <- parseCVs as
+                   ((b, c, d), es) <- parseCVs bs
                    return ((a, b, c, d), es)
 
 -- | 5-Tuples extracted from a model
 instance (SatModel a, SatModel b, SatModel c, SatModel d, SatModel e) => SatModel (a, b, c, d, e) where
-  parseCWs as = do (a, bs)            <- parseCWs as
-                   ((b, c, d, e), fs) <- parseCWs bs
+  parseCVs as = do (a, bs)            <- parseCVs as
+                   ((b, c, d, e), fs) <- parseCVs bs
                    return ((a, b, c, d, e), fs)
 
 -- | 6-Tuples extracted from a model
 instance (SatModel a, SatModel b, SatModel c, SatModel d, SatModel e, SatModel f) => SatModel (a, b, c, d, e, f) where
-  parseCWs as = do (a, bs)               <- parseCWs as
-                   ((b, c, d, e, f), gs) <- parseCWs bs
+  parseCVs as = do (a, bs)               <- parseCVs as
+                   ((b, c, d, e, f), gs) <- parseCVs bs
                    return ((a, b, c, d, e, f), gs)
 
 -- | 7-Tuples extracted from a model
 instance (SatModel a, SatModel b, SatModel c, SatModel d, SatModel e, SatModel f, SatModel g) => SatModel (a, b, c, d, e, f, g) where
-  parseCWs as = do (a, bs)                  <- parseCWs as
-                   ((b, c, d, e, f, g), hs) <- parseCWs bs
+  parseCVs as = do (a, bs)                  <- parseCVs as
+                   ((b, c, d, e, f, g), hs) <- parseCVs bs
                    return ((a, b, c, d, e, f, g), hs)
 
 -- | Various SMT results that we can extract models out of.
@@ -322,19 +332,19 @@ class Modelable a where
 
   -- | Extract a model dictionary. Extract a dictionary mapping the variables to
   -- their respective values as returned by the SMT solver. Also see `getModelDictionaries`.
-  getModelDictionary :: a -> M.Map String CW
+  getModelDictionary :: a -> M.Map String CV
 
   -- | Extract a model value for a given element. Also see `getModelValues`.
-  getModelValue :: SymWord b => String -> a -> Maybe b
-  getModelValue v r = fromCW `fmap` (v `M.lookup` getModelDictionary r)
+  getModelValue :: SymVal b => String -> a -> Maybe b
+  getModelValue v r = fromCV `fmap` (v `M.lookup` getModelDictionary r)
 
   -- | Extract a representative name for the model value of an uninterpreted kind.
   -- This is supposed to correspond to the value as computed internally by the
   -- SMT solver; and is unportable from solver to solver. Also see `getModelUninterpretedValues`.
   getModelUninterpretedValue :: String -> a -> Maybe String
   getModelUninterpretedValue v r = case v `M.lookup` getModelDictionary r of
-                                     Just (CW _ (CWUserSort (_, s))) -> Just s
-                                     _                               -> Nothing
+                                     Just (CV _ (CUserSort (_, s))) -> Just s
+                                     _                              -> Nothing
 
   -- | A simpler variant of 'getModelAssignment' to get a model out without the fuss.
   extractModel :: SatModel b => a -> Maybe b
@@ -343,50 +353,59 @@ class Modelable a where
                      _            -> Nothing
 
   -- | Extract model objective values, for all optimization goals.
-  getModelObjectives :: a -> M.Map String GeneralizedCW
+  getModelObjectives :: a -> M.Map String GeneralizedCV
 
   -- | Extract the value of an objective
-  getModelObjectiveValue :: String -> a -> Maybe GeneralizedCW
+  getModelObjectiveValue :: String -> a -> Maybe GeneralizedCV
   getModelObjectiveValue v r = v `M.lookup` getModelObjectives r
 
+  -- | Extract model uninterpreted-functions
+  getModelUIFuns :: a -> M.Map String (SBVType, ([([CV], CV)], CV))
+
+  -- | Extract the value of an uninterpreted-function as an association list
+  getModelUIFunValue :: String -> a -> Maybe (SBVType, ([([CV], CV)], CV))
+  getModelUIFunValue v r = v `M.lookup` getModelUIFuns r
+
 -- | Return all the models from an 'Data.SBV.allSat' call, similar to 'extractModel' but
 -- is suitable for the case of multiple results.
 extractModels :: SatModel a => AllSatResult -> [a]
-extractModels (AllSatResult (_, _, xs)) = [ms | Right (_, ms) <- map getModelAssignment xs]
+extractModels (AllSatResult (_, _, _, xs)) = [ms | Right (_, ms) <- map getModelAssignment xs]
 
 -- | Get dictionaries from an all-sat call. Similar to `getModelDictionary`.
-getModelDictionaries :: AllSatResult -> [M.Map String CW]
-getModelDictionaries (AllSatResult (_, _, xs)) = map getModelDictionary xs
+getModelDictionaries :: AllSatResult -> [M.Map String CV]
+getModelDictionaries (AllSatResult (_, _, _, xs)) = map getModelDictionary xs
 
 -- | Extract value of a variable from an all-sat call. Similar to `getModelValue`.
-getModelValues :: SymWord b => String -> AllSatResult -> [Maybe b]
-getModelValues s (AllSatResult (_, _, xs)) =  map (s `getModelValue`) xs
+getModelValues :: SymVal b => String -> AllSatResult -> [Maybe b]
+getModelValues s (AllSatResult (_, _, _, xs)) =  map (s `getModelValue`) xs
 
 -- | Extract value of an uninterpreted variable from an all-sat call. Similar to `getModelUninterpretedValue`.
 getModelUninterpretedValues :: String -> AllSatResult -> [Maybe String]
-getModelUninterpretedValues s (AllSatResult (_, _, xs)) =  map (s `getModelUninterpretedValue`) xs
+getModelUninterpretedValues s (AllSatResult (_, _, _, xs)) =  map (s `getModelUninterpretedValue`) xs
 
 -- | 'ThmResult' as a generic model provider
 instance Modelable ThmResult where
   getModelAssignment (ThmResult r) = getModelAssignment r
-  modelExists        (ThmResult r) = modelExists r
+  modelExists        (ThmResult r) = modelExists        r
   getModelDictionary (ThmResult r) = getModelDictionary r
   getModelObjectives (ThmResult r) = getModelObjectives r
+  getModelUIFuns     (ThmResult r) = getModelUIFuns     r
 
 -- | 'SatResult' as a generic model provider
 instance Modelable SatResult where
   getModelAssignment (SatResult r) = getModelAssignment r
-  modelExists        (SatResult r) = modelExists r
+  modelExists        (SatResult r) = modelExists        r
   getModelDictionary (SatResult r) = getModelDictionary r
   getModelObjectives (SatResult r) = getModelObjectives r
+  getModelUIFuns     (SatResult r) = getModelUIFuns     r
 
 -- | 'SMTResult' as a generic model provider
 instance Modelable SMTResult where
-  getModelAssignment (Unsatisfiable _ _) = Left "SBV.getModelAssignment: Unsatisfiable result"
-  getModelAssignment (Satisfiable _ m)   = Right (False, parseModelOut m)
-  getModelAssignment (SatExtField _ _)   = Left "SBV.getModelAssignment: The model is in an extension field"
-  getModelAssignment (Unknown _ m)       = Left $ "SBV.getModelAssignment: Solver state is unknown: " ++ show m
-  getModelAssignment (ProofError _ s)    = error $ unlines $ "Backend solver complains: " : s
+  getModelAssignment (Unsatisfiable _ _  ) = Left "SBV.getModelAssignment: Unsatisfiable result"
+  getModelAssignment (Satisfiable   _ m  ) = Right (False, parseModelOut m)
+  getModelAssignment (SatExtField   _ _  ) = Left "SBV.getModelAssignment: The model is in an extension field"
+  getModelAssignment (Unknown       _ m  ) = Left $ "SBV.getModelAssignment: Solver state is unknown: " ++ show m
+  getModelAssignment (ProofError    _ s _) = error $ unlines $ "SBV.getModelAssignment: Failed to produce a model: " : s
 
   modelExists Satisfiable{}   = True
   modelExists Unknown{}       = False -- don't risk it
@@ -404,9 +423,15 @@ instance Modelable SMTResult where
   getModelObjectives Unknown{}         = M.empty
   getModelObjectives ProofError{}      = M.empty
 
+  getModelUIFuns Unsatisfiable{}   = M.empty
+  getModelUIFuns (Satisfiable _ m) = M.fromList (modelUIFuns m)
+  getModelUIFuns (SatExtField _ m) = M.fromList (modelUIFuns m)
+  getModelUIFuns Unknown{}         = M.empty
+  getModelUIFuns ProofError{}      = M.empty
+
 -- | Extract a model out, will throw error if parsing is unsuccessful
 parseModelOut :: SatModel a => SMTModel -> a
-parseModelOut m = case parseCWs [c | (_, c) <- modelAssocs m] of
+parseModelOut m = case parseCVs [c | (_, c) <- modelAssocs m] of
                    Just (x, []) -> x
                    Just (_, ys) -> error $ "SBV.parseModelOut: Partially constructed model; remaining elements: " ++ show ys
                    Nothing      -> error $ "SBV.parseModelOut: Cannot construct a model from: " ++ show m
@@ -414,23 +439,32 @@ parseModelOut m = case parseCWs [c | (_, c) <- modelAssocs m] of
 -- | Given an 'Data.SBV.allSat' call, we typically want to iterate over it and print the results in sequence. The
 -- 'displayModels' function automates this task by calling @disp@ on each result, consecutively. The first
 -- 'Int' argument to @disp@ 'is the current model number. The second argument is a tuple, where the first
--- element indicates whether the model is alleged (i.e., if the solver is not sure, returing Unknown)
-displayModels :: SatModel a => (Int -> (Bool, a) -> IO ()) -> AllSatResult -> IO Int
-displayModels disp (AllSatResult (_, _, ms)) = do
-    inds <- zipWithM display [a | Right a <- map (getModelAssignment . SatResult) ms] [(1::Int)..]
+-- element indicates whether the model is alleged (i.e., if the solver is not sure, returing Unknown).
+-- The arrange argument can sort the results in any way you like, if necessary.
+displayModels :: SatModel a => ([(Bool, a)] -> [(Bool, a)]) -> (Int -> (Bool, a) -> IO ()) -> AllSatResult -> IO Int
+displayModels arrange disp (AllSatResult (_, _, _, ms)) = do
+    let models = [a | Right a <- map (getModelAssignment . SatResult) ms]
+    inds <- zipWithM display (arrange models) [(1::Int)..]
     return $ last (0:inds)
   where display r i = disp i r >> return i
 
 -- | Show an SMTResult; generic version
 showSMTResult :: String -> String -> String -> String -> String -> SMTResult -> String
 showSMTResult unsatMsg unkMsg satMsg satMsgModel satExtMsg result = case result of
-  Unsatisfiable _ uc            -> unsatMsg ++ showUnsatCore uc
-  Satisfiable _ (SMTModel _ []) -> satMsg
-  Satisfiable _ m               -> satMsgModel ++ showModel cfg m
-  SatExtField _ (SMTModel b _)  -> satExtMsg   ++ showModelDictionary cfg b
-  Unknown     _ r               -> unkMsg ++ ".\n" ++ "  Reason: " `alignPlain` show r
-  ProofError  _ []              -> "*** An error occurred. No additional information available. Try running in verbose mode"
-  ProofError  _ ls              -> "*** An error occurred.\n" ++ intercalate "\n" (map ("***  " ++) ls)
+  Unsatisfiable _ uc                 -> unsatMsg ++ showUnsatCore uc
+  Satisfiable _ (SMTModel _ _ [] []) -> satMsg
+  Satisfiable _ m                    -> satMsgModel ++ showModel cfg m
+  SatExtField _ (SMTModel b _ _ _)   -> satExtMsg   ++ showModelDictionary True False cfg b
+  Unknown     _ r                    -> unkMsg ++ ".\n" ++ "  Reason: " `alignPlain` show r
+  ProofError  _ [] Nothing           -> "*** An error occurred. No additional information available. Try running in verbose mode."
+  ProofError  _ ls Nothing           -> "*** An error occurred.\n" ++ intercalate "\n" (map ("***  " ++) ls)
+  ProofError  _ ls (Just r)          -> intercalate "\n" $  [ "*** " ++ l | l <- ls]
+                                                         ++ [ "***"
+                                                            , "*** Alleged model:"
+                                                            , "***"
+                                                            ]
+                                                         ++ ["*** "  ++ l | l <- lines (showSMTResult unsatMsg unkMsg satMsg satMsgModel satExtMsg r)]
+
  where cfg = resultConfig result
        showUnsatCore Nothing   = ""
        showUnsatCore (Just xs) = ". Unsat core:\n" ++ intercalate "\n" ["    " ++ x | x <- xs]
@@ -438,21 +472,33 @@ showSMTResult unsatMsg unkMsg satMsg satMsgModel satExtMsg result = case result
 -- | Show a model in human readable form. Ignore bindings to those variables that start
 -- with "__internal_sbv_" and also those marked as "nonModelVar" in the config; as these are only for internal purposes
 showModel :: SMTConfig -> SMTModel -> String
-showModel cfg model = showModelDictionary cfg [(n, RegularCW c) | (n, c) <- modelAssocs model]
+showModel cfg model
+   | null uiFuncs
+   = nonUIFuncs
+   | True
+   = sep nonUIFuncs ++ intercalate "\n\n" (map (showModelUI cfg) uiFuncs)
+   where nonUIFuncs = showModelDictionary (null uiFuncs) False cfg [(n, RegularCV c) | (n, c) <- modelAssocs model]
+         uiFuncs    = modelUIFuns model
+         sep ""     = ""
+         sep x      = x ++ "\n\n"
 
 -- | Show bindings in a generalized model dictionary, tabulated
-showModelDictionary :: SMTConfig -> [(String, GeneralizedCW)] -> String
-showModelDictionary cfg allVars
+showModelDictionary :: Bool -> Bool -> SMTConfig -> [(String, GeneralizedCV)] -> String
+showModelDictionary warnEmpty includeEverything cfg allVars
    | null allVars
-   = "[There are no variables bound by the model.]"
+   = warn "[There are no variables bound by the model.]"
    | null relevantVars
-   = "[There are no model-variables bound by the model.]"
+   = warn "[There are no model-variables bound by the model.]"
    | True
    = intercalate "\n" . display . map shM $ relevantVars
-  where relevantVars  = filter (not . ignore) allVars
-        ignore (s, _) = "__internal_sbv_" `isPrefixOf` s || isNonModelVar cfg s
+  where warn s = if warnEmpty then s else ""
 
-        shM (s, RegularCW v) = let vs = shCW cfg v in ((length s, s), (vlength vs, vs))
+        relevantVars  = filter (not . ignore) allVars
+        ignore (s, _)
+          | includeEverything = False
+          | True              = "__internal_sbv_" `isPrefixOf` s || isNonModelVar cfg s
+
+        shM (s, RegularCV v) = let vs = shCV cfg v in ((length s, s), (vlength vs, vs))
         shM (s, other)       = let vs = show other in ((length s, s), (vlength vs, vs))
 
         display svs   = map line svs
@@ -472,9 +518,48 @@ showModelDictionary cfg allVars
 
         lTrimRight = length . dropWhile isSpace . reverse
 
+-- | Show an uninterpreted function
+showModelUI :: SMTConfig -> (String, (SBVType, ([([CV], CV)], CV))) -> String
+showModelUI cfg (nm, (SBVType ts, (defs, dflt))) = intercalate "\n" ["  " ++ l | l <- sig : map align body]
+  where noOfArgs = length ts - 1
+
+        sig      = nm ++ " :: " ++ intercalate " -> " (map showBaseKind ts)
+
+        ls       = map line defs
+        defLine  = (nm : replicate noOfArgs "_", scv dflt)
+
+        body     = ls ++ [defLine]
+
+        colWidths = [maximum (0 : map length col) | col <- transpose (map fst body)]
+
+        resWidth  = maximum  (0 : map (length . snd) body)
+
+        align (xs, r) = unwords $ zipWith left colWidths xs ++ ["=", left resWidth r]
+           where left i x = take i (x ++ repeat ' ')
+
+        scv = sh (printBase cfg)
+          where sh 2  = binP
+                sh 10 = showCV False
+                sh 16 = hexP
+                sh _  = show
+
+        -- NB. If we have a float NaN/Infinity/+0/-0 etc. these will
+        -- simply print as is, but will not be valid patterns. (We
+        -- have the semi-goal of being able to paste these definitions
+        -- in a Haskell file.) For the time being, punt on this, but
+        -- we might want to do this properly later somehow. (Perhaps
+        -- using some sort of a view pattern.)
+        line :: ([CV], CV) -> ([String], String)
+        line (args, r) = (nm : map (paren . scv) args, scv r)
+          where -- If negative, parenthesize. I think this is the only case
+                -- we need to worry about. Hopefully!
+                paren :: String -> String
+                paren x@('-':_) = '(' : x ++ ")"
+                paren x         = x
+
 -- | Show a constant value, in the user-specified base
-shCW :: SMTConfig -> CW -> String
-shCW = sh . printBase
+shCV :: SMTConfig -> CV -> String
+shCV = sh . printBase
   where sh 2  = binS
         sh 10 = show
         sh 16 = hexS
@@ -538,25 +623,15 @@ runSolver cfg ctx execPath opts pgm continuation
  = do let nm  = show (name (solver cfg))
           msg = debug cfg . map ("*** " ++)
 
+          clean = preprocess (solver cfg)
+
       (send, ask, getResponseFromSolver, terminateSolver, cleanUp, pid) <- do
                 (inh, outh, errh, pid) <- runInteractiveProcess execPath opts Nothing Nothing
 
-                let -- send a command down, but check that we're balanced in parens. If we aren't
-                    -- this is most likely an SBV bug.
-                    send :: Maybe Int -> String -> IO ()
-                    send mbTimeOut command
-                      | parenDeficit command /= 0
-                      = error $ unlines [ ""
-                                        , "*** Data.SBV: Unbalanced input detected."
-                                        , "***"
-                                        , "***   Sending: " ++ command
-                                        , "***"
-                                        , "*** This is most likely an SBV bug. Please report!"
-                                        ]
-                      | True
-                      = do hPutStrLn inh command
-                           hFlush inh
-                           recordTranscript (transcript cfg) $ Left (command, mbTimeOut)
+                let send :: Maybe Int -> String -> IO ()
+                    send mbTimeOut command = do hPutStrLn inh (clean command)
+                                                hFlush inh
+                                                recordTranscript (transcript cfg) $ Left (command, mbTimeOut)
 
                     -- Send a line, get a whole s-expr. We ignore the pathetic case that there might be a string with an unbalanced parentheses in it in a response.
                     ask :: Maybe Int -> String -> IO String
@@ -583,8 +658,8 @@ runSolver cfg ctx execPath opts pgm continuation
                       where safeGetLine isFirst h =
                                          let timeOutToUse | isFirst = mbTimeOut
                                                           | True    = Just 5000000
-                                             timeOutMsg t | isFirst = "User specified timeout of " ++ showTimeoutValue t ++ " exceeded."
-                                                          | True    = "A multiline complete response wasn't received before " ++ showTimeoutValue t ++ " exceeded."
+                                             timeOutMsg t | isFirst = "User specified timeout of " ++ showTimeoutValue t ++ " exceeded"
+                                                          | True    = "A multiline complete response wasn't received before " ++ showTimeoutValue t ++ " exceeded"
 
                                              -- Like hGetLine, except it keeps getting lines if inside a string.
                                              getFullLine :: IO String
@@ -676,6 +751,9 @@ runSolver cfg ctx execPath opts pgm continuation
                                 ++ [ "Std-out  : " ++ intercalate "\n           " (lines out) | not (null out)]
                                 ++ [ "Std-err  : " ++ intercalate "\n           " (lines err) | not (null err)]
 
+                           finalizeTranscript (transcript cfg) ex
+                           recordEndTime cfg ctx
+
                            case ex of
                              ExitSuccess -> return ()
                              _           -> if ignoreExitCode cfg
@@ -796,7 +874,7 @@ runSolver cfg ctx execPath opts pgm continuation
                                                  , queryAssertionStackDepth = 0
                                                  , queryTblArrPreserveIndex = Nothing
                                                  }
-                                 qsp = queryState ctx
+                                 qsp = rQueryState ctx
 
                              mbQS <- readIORef qsp
 
@@ -811,17 +889,14 @@ runSolver cfg ctx execPath opts pgm continuation
                              continuation ctx
 
       -- NB. Don't use 'bracket' here, as it wouldn't have access to the exception.
-      let launchSolver = do startTranscript    (transcript cfg) cfg
-                            r <- executeSolver
-                            finalizeTranscript (transcript cfg) Nothing
-                            recordEndTime      cfg ctx
-                            return r
+      let launchSolver = do startTranscript (transcript cfg) cfg
+                            executeSolver
 
       launchSolver `C.catch` (\(e :: C.SomeException) -> handleAsync e $ do terminateProcess pid
                                                                             ec <- waitForProcess pid
                                                                             recordException    (transcript cfg) (show e)
-                                                                            finalizeTranscript (transcript cfg) (Just ec)
-                                                                            recordEndTime      cfg ctx
+                                                                            finalizeTranscript (transcript cfg) ec
+                                                                            recordEndTime cfg ctx
                                                                             C.throwIO e)
 
 -- | Compute and report the end time
@@ -853,19 +928,19 @@ startTranscript (Just f) cfg = do ts <- show <$> getZonedTime
                                   ]
 
 -- | Finish up the transcript file.
-finalizeTranscript :: Maybe FilePath -> Maybe ExitCode -> IO ()
-finalizeTranscript Nothing  _    = return ()
-finalizeTranscript (Just f) mbEC = do ts <- show <$> getZonedTime
-                                      appendFile f $ end ts
-  where end ts = unlines $ [ ""
-                           , ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
-                           , ";;;"
-                           , ";;; SBV: Finished at " ++ ts
-                           ]
-                       ++  [ ";;;\n;;; Exit code: " ++ show ec | Just ec <- [mbEC] ]
-                       ++  [ ";;;"
-                           , ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
-                           ]
+finalizeTranscript :: Maybe FilePath -> ExitCode -> IO ()
+finalizeTranscript Nothing  _  = return ()
+finalizeTranscript (Just f) ec = do ts <- show <$> getZonedTime
+                                    appendFile f $ end ts
+  where end ts = unlines [ ""
+                         , ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
+                         , ";;;"
+                         , ";;; SBV: Finished at " ++ ts
+                         , ";;;"
+                         , ";;; Exit code: " ++ show ec
+                         , ";;;"
+                         , ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
+                         ]
 
 -- If requested, record in the transcript file
 recordTranscript :: Maybe FilePath -> Either (String, Maybe Int) String -> IO ()
